(* Skip lists for ocaml
   Copyright (C) 2002 Shawn Wagner <raevnos@pennmush.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
	 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
	 
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

module Array = ArrayLabels

module type OrderedType =
sig
  type t
  val compare: t -> t -> int
end
	
module type S =
sig
  type key
  type 'a t
  val create : ?size:int -> unit -> 'a t
  val clear : 'a t -> unit
  val size : 'a t -> int
  val add : 'a t -> key:key -> data:'a -> unit
  val find : 'a t -> key -> 'a
  val remove : 'a t -> key -> unit 
  val mem : 'a t -> key -> bool
  val iter : 'a t -> f:(key -> 'a -> unit) -> unit
  val fold : 'a t -> f:(key -> 'a -> 'b -> 'b)
    -> init:'b -> 'b
end

module RNG = Rand.UniformDist(Math.Int32Ops)(MtRand.Int32Source)
let rng = new RNG.rng Int32.zero (Int32.pred Int32.max_int)
let genrand () = Int32.to_float rng#genrand

module Make(Ord: OrderedType): (S with type key = Ord.t) =
struct
  let mi_as_float = float_of_int max_int

  (* Use t == e, for best theoretical performance *)
  let private_rand size =
    let t = genrand () in
    let rec build_i i j =
      if i >= size or t > mi_as_float /. j then
	min i size
      else 
	build_i (i + 1) (j +. j) in
      build_i 1 Math.e
	
  type key = Ord.t
  and 'a skip_t = SkipNode of key * 'a * 'a skip_t array
	          | Empty
  and 'a t = {
    max_size : int;
    count : int ref;
    first_chain : 'a skip_t array;
  }
	       
  let get_key node = 
    match node with
      |	SkipNode(key, _, _) -> key
      |	Empty -> raise Not_found

  let get_size node =
    match node with
      | SkipNode(_, _, chain) -> Array.length chain
      |	Empty -> 0
	  
  let get_chain node =
    match node with
      |	SkipNode(_, _, chain) -> chain
      |	 Empty -> raise Not_found
	   
  let create ?(size=5) () = 
    let chainsize = max size 2 in
      {
        max_size = chainsize;
        count = ref 0;
        first_chain = Array.create chainsize Empty;
      }

  let clear sl =
    sl.count := 0;
    Array.fill sl.first_chain ~pos:0 ~len:sl.max_size Empty
      
  let size sl = !(sl.count)
		  
  let add sl ~key ~data =
    let rec insertR curr newnode pos =
      if pos < 0 then () else begin
	match curr.(pos) with
	  | Empty -> if pos < get_size newnode then
	      begin
		Array.set (get_chain newnode) pos curr.(pos);
		Array.set curr pos newnode
	      end;
	      insertR curr newnode (pos - 1)
	  |	SkipNode(ky, _, chain) -> begin
		  match Ord.compare (get_key newnode) ky with
		    |	x when x < 0 -> if pos < get_size newnode then
			  begin
			    Array.set (get_chain newnode) pos curr.(pos);
			    Array.set curr pos newnode
			  end;
			  insertR curr newnode (pos - 1)
		    |	_ -> insertR chain newnode pos
		end 
      end in
    let thissize = private_rand (sl.max_size - 1) in
      incr sl.count;
      insertR sl.first_chain
	(SkipNode(key, data, Array.create thissize Empty))
	(sl.max_size - 1)

  let find sl key = 
    let rec searchR curr pos =
      if pos < 0 then raise Not_found;
      match curr.(pos) with
	| Empty -> searchR curr (pos - 1)
	|	SkipNode(ky, data, chain) -> begin
		  match Ord.compare key ky with
		    |	0 -> data
		    |	x when x < 0 -> searchR curr (pos - 1)
		    |	_ -> searchR chain pos
		end in
      searchR sl.first_chain (sl.max_size - 1)
	
  let remove sl key =
    let found = ref 0 in
    let rec deleteR chain pos =
      if pos < 0 then () else
	match chain.(pos) with
	  |	Empty -> deleteR chain (pos - 1)
	  |	SkipNode(ky, _, next) ->
		  begin
		    match Ord.compare ky key with
		      |	x when x < 0 ->  deleteR next pos 
		      | 0 ->
			  Array.set chain pos next.(pos);
			  incr found;
			  deleteR chain (pos - 1)
		      |	_ -> deleteR chain (pos - 1)
		  end
    in
    let _ = deleteR sl.first_chain (sl.max_size - 1) in
      if !found > 0 then decr sl.count
	
	let mem sl key =
	  try
	    let _ = find sl key in
	      true
	  with
	      Not_found -> false
		
	let iter sl ~f =
	  let rec apply node =
	    match node with
	      |	Empty -> ()
	      |	SkipNode(key, data, chain) -> 
		  f key data;
		  apply chain.(0) in
	    apply sl.first_chain.(0)


	let fold sl ~f ~init =
	  let rec apply node arg =
	    match node with
	      |	Empty -> arg
	      |	SkipNode(key, data, chain) ->
		  apply chain.(0) (f key data arg) in
	    apply sl.first_chain.(0) init
	      
end

