(* Skip lists for ocaml
   Copyright (C) 2002 Shawn Wagner <raevnos@pennmush.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
	 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
	 
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)


(** Module that provides the skiplist datastructure. *)

(** Types used as skiplist keys have to be ordered *)
module type OrderedType =
  sig
    type t
    val compare: t -> t -> int
  end

(** The signature of skiplists for a given key type *)
module type S =
  sig
    (** The key type *)
    type key
      
    (** The type of skiplists *)
    type 'a t

      (** Create a new skiplist with given max height, or 5 if none is given. *)
    val create : ?size:int -> unit -> 'a t

    (** Wipe out the skiplist *)
    val clear : 'a t -> unit
    
    (** The number of elements in the skiplist *)
    val size : 'a t -> int
    
    (** Insert an element *)
    val add : 'a t -> key:key -> data:'a -> unit
      
    (** Look up an element *)
    val find : 'a t -> key -> 'a

    (** Remove an element *)
    val remove : 'a t -> key -> unit

    (** Like [List.mem] *)
    val mem : 'a t -> key -> bool

    (** Like [Map.iter] *)
    val iter : 'a t -> f:(key -> 'a -> unit) -> unit

    (** Like [Map.fold] *)
    val fold : 'a t -> f:(key -> 'a -> 'b -> 'b) -> init:'b -> 'b
  end

(** Create a new skiplist module for the given key type. *)
module Make(Ord: OrderedType): (S with type key = Ord.t)
