(* string  utility routines for ocaml
   Copyright (C) 2002 Shawn Wagner <raevnos@pennmush.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
	 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
	 
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

module String = StdLabels.String

(* Returns the string minus trailing whitespace *)
let chomp str =
  let c = ref true in
  let p = ref ((String.length str)) in
    while !c do
      if !p <= 0 then begin
        p := 0; c := false; end
      else begin
        if CharExtras.is_space (String.get str (!p - 1)) then decr p
        else c := false end
    done;
    String.sub str ~pos:0 ~len:!p
      

(* Returns the first word of a string *)
let first_word str =
	try
		let len = String.index_from str 1 ' ' in
		  String.sub str ~pos:0 ~len
	with
		  Not_found -> str

(* Returns everything after the character. *)
let split_at ~str ~sep  =
	try
		let p = String.index str sep in
		let len = String.length str - p - 1 
		and pos = succ p in
		  String.sub str ~pos ~len
	with
		  Not_found -> ""

(* Cuts off the first character of a string and returns the rest *)
let cut_first_char str =
	let slen = String.length str in
	let len = pred slen in 
	  String.sub str ~pos:1 ~len

(* Cuts off the last character of a string and returns the rest *)
let cut_last_char str = 
	let len = String.length str - 1 in
	  String.sub str ~pos:0 ~len

(* Cuts off the first word of a string and returns the rest *)
let rec cut_first_word str =
	let rest = split_at ~str ~sep:' ' in
	  try
		  let rest_cut = cut_first_char rest in 
		    if String.sub rest ~pos:0 ~len:1 = " " then
			    cut_first_word rest_cut
		    else
			    rest
	  with
		    Not_found -> rest

(* Concatenates a list of strings with a null seperator *)
let combine lst =
	String.concat ~sep:"" lst


(* Return str minus the first n characters *)
let cut_first_n str n =
  let len = String.length str in
    if n > len then
      ""
    else
      String.sub str n (len - n)

(* Return str minus the last n characters *)
let cut_last_n str n = 
  let len = String.length str in
    if n > len then
      ""
    else
      String.sub str 0 (len - n)


let make_search_vec needle def =
  let v = Array.make 255 def in
    for n = 0 to (String.length needle) - 1 do
      v.(Char.code needle.[n]) <- not def
    done;
    v

let rec checkt2 vec hay pos =
  if vec.(Char.code hay.[pos]) then
    pos
  else
    checkt2 vec hay (pos + 1)

let checkt vec hay pos =
  try
    checkt2 vec hay pos
  with
    | Invalid_argument _ -> raise Not_found

let first_of_from needle haystack pos =
  let need = make_search_vec needle false in
    checkt need haystack pos
    
let first_not_of_from needle haystack pos =
  let need = make_search_vec needle true in
    checkt need haystack pos

let first_of needle haystack = first_of_from needle haystack 0
let first_not_of needle haystack = first_not_of_from needle haystack 0

let map f str =
  let len = String.length str in
  let nstr = String.create len in
    for n = 0 to len - 1 do
      nstr.[n] <- f str.[n]
    done;
    nstr
    
let map_inplace f str =
  let len = String.length str in
    for n = 0 to len - 1 do
      str.[n] <- f str.[n]
    done

let uppercase = map CharExtras.to_upper
let lowercase = map CharExtras.to_lower

let capitalize str = 
  if String.length str = 0 then ""
  else
    let nstr = String.copy str in
      nstr.[0] <- CharExtras.to_upper nstr.[0];
      nstr

let uncapitalize str = 
  if String.length str = 0 then ""
  else
    let nstr = String.copy str in
      nstr.[0] <- CharExtras.to_lower nstr.[0];
      nstr

let prefix pref str =
  let plen = String.length pref in
  if plen > String.length str then false
  else if plen = String.length str then compare pref str = 0
  else let p = ref 0 in
    while !p < plen && pref.[!p] = str.[!p] do
      incr p
    done;
    !p = plen

let suffix suf str =
  let slen = String.length suf 
  and len = String.length str in
  if slen > len then false
  else if slen = len then compare suf str = 0
  else let p = ref (slen - 1) 
       and n = ref (len - 1) in
    while !p >= 0 && suf.[!p] = str.[!n] do
      decr p;
      decr n;
    done;
    !p = -1

let rec find_str find flen str slen pos =
  let fp = ref 0 
  and sp = ref pos in
    while !fp < flen && !sp < slen && find.[!fp] = str.[!sp] do
      incr fp;
      incr sp;
    done;
    if !fp = flen then
      pos
    else
      find_str find flen str slen (String.index_from str (pos + 1) find.[0])

let index_substr_from find str pos =
  if find = "" then 0
  else if str = "" then raise Not_found
  else
    find_str find (String.length find) str (String.length str)
      (String.index_from str pos find.[0])
	

let index_substr find str = index_substr_from find str 0  


let match_substr find str pos =
  let slen = String.length str in
    if pos < 0 || pos >= slen then
      raise (Invalid_argument "StrExtras.match_substr")
    else
      let fp = ref 0
      and flen = String.length find 
      and sp = ref pos in
	while !fp < flen && !sp < slen && find.[!fp] = str.[!sp] do
	  incr fp;
	  incr sp;
	done;
	!fp = flen
