(* string  utility routines for ocaml
   Copyright (C) 2002 Shawn Wagner <raevnos@pennmush.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
	 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
	 
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(** String functions *)

(** {1 String construction} *)

(** Concatenates a list of strings with a blank seperator *)
val combine: string list -> string

(** Like [Array.map], for strings. Returns a newly allocated transformed string.  *)
val map: (char -> char) -> string -> string

(** {2 String manipulation} *)

(** Most of these functions return newly allocated strings that are modified versions of their arguments *)

(** Returns the first word of a string *)
val first_word: string -> string

(** Cuts off the first character of a string and returns the rest *)
val cut_first_char: string -> string

(** Cuts off the first [n] characters of a string and returns the rest. *)
val cut_first_n: string -> int -> string

(** Cuts off the last character of a string and returns the rest *)
val cut_last_char: string -> string

(** Cuts off the last [n] characters of a string and returns the rest *)
val cut_last_n: string -> int -> string

(** Cuts off the first word of a string and returns the rest *)
val cut_first_word: string -> string

(** Returns everything after the character *)
val split_at: str:string -> sep:char -> string

(** Remove all trailing whitespace *)
val chomp: string -> string

(** Like [map], but modifies the argument string. *)
val map_inplace: (char -> char) -> string -> unit

(** {3 Capitalization} *)

(** Like [String.uppercase], but locale-dependant *)
val uppercase: string -> string

(** Like [String.lowercase], but locale-dependant *)
val lowercase: string -> string

(** Like [String.capitalize], but locale-dependant *)
val capitalize: string -> string

(** Like [String.uncapitalize], but locale-dependant *)
val uncapitalize: string -> string


(** {1 String searching} *)

(** [first_of needle haystack] returns the position of the first occurance in haystack of a character in needle. Like C strcspn(). *)
val first_of: string -> string -> int
(**
@raise Not_found if no characters in needle are in haystack 
*)

(** [first_of_from needle haystack pos] returns the position of the first occurance in haystack (Starting at pos) of a character in needle. *)
val first_of_from: string -> string -> int -> int
(**
@raise Not_found if no characters in needle are in haystack 
*)

(** [first_not_of needle haystack] returns the position of the first occurance in haystack of a character that's not also in needle. Like C strspn(). *)
val first_not_of: string -> string -> int
(**
@raise Not_found if all characters in needle are in haystack 
*)

(** [first_not_of_from needle haystack pos] returns the position of the first occurance in haystack (Starting at pos) of a character that's not also in needle. *)
val first_not_of_from: string -> string -> int -> int
(**
@raise Not_found if all characters in needle are in haystack 
*)

(** [prefix pref str] returns true if str starts with pref. *)
val prefix: string -> string -> bool

(** [suffix suf str] returns true if str starts with suf. *)
val suffix: string -> string -> bool

(** [index_str needle haystack] returns the position in haystack where needle starts. *)
val index_substr: string -> string -> int
(**
@raise Not_found if the substring isn't present.
*)

(** [index_substr_from needle haystack pos] returns the position in haystack where needle starts, starting looking at pos. *)
val index_substr_from: string -> string -> int -> int
(**
@raise Not_found if the substring isn't present.
@raise Invalid_argument if the positition is outside of haystack.
*)

(** [match_substr substr str pos] returns true if str contains substr at position pos *)
val match_substr: string -> string -> int -> bool
(**
@raise Invalid_argument if pos is out of range 
*)

